import ctypes

from kydns.kyd_exc import DNSInvalidDomain
from kydns.protocol import PPrinter

MAX_LABEL_LENGTH = 63


class QCLASS:
    IN = 1


class QTYPE:
    A = 1
    NS = 2
    AAAA = 28


class DNSDomain:
    def __init__(self, domain, offset: bool = False):
        self.domain = domain.lower().rstrip('.') if isinstance(domain, str) else domain
        self.offset = offset
        self.validate()

    def __len__(self):
        return len(self.__bytes__())

    def __repr__(self):
        return self.domain

    def __bytes__(self) -> bytes:
        if self.offset:
            return bytes.fromhex("c00c")
        encoded = []
        for label in self.domain.split('.'):
            enc_label = label.encode('utf-8')
            enc_len = len(enc_label).to_bytes(1, byteorder="big")
            encoded.append(enc_len + enc_label)
        return b''.join(encoded) + b'\x00'

    def validate(self) -> None:
        for label in self.domain.split('.'):
            if len(label) > MAX_LABEL_LENGTH:
                raise DNSInvalidDomain(f"Label too long: '{label}'")

    @classmethod
    def to_domain(cls, data: bytes, index: int = 0) -> 'DNSDomain':
        data_len = len(data)
        offset = False
        domain = ""

        while index < data_len:
            label_len = data[index]
            if not label_len:
                break
            elif (data[index] >> 6) == 3:
                offset = True
                index = int.from_bytes(data[index:index + 2], 'big') & 0x3f
                continue
            domain += data[index + 1:index + 1 + label_len].decode('utf-8') + "."
            index += (label_len + 1)
        return cls(domain, offset)


class DNSHeader(ctypes.BigEndianStructure):
    _fields_ = [
        ("id", ctypes.c_uint16),        # 16bit identifier
        ("qr", ctypes.c_uint8, 1),      # 1bit 0-query 1-response
        ("opcode", ctypes.c_uint8, 4),  # 4bit query type 0-standard
        ("aa", ctypes.c_uint8, 1),      # 1bit authorative answer
        ("tc", ctypes.c_uint8, 1),      # 1bit truncation
        ("rd", ctypes.c_uint8, 1),      # 1bit recursion desired
        ("ra", ctypes.c_uint8, 1),      # 1bit recursion available
        ("z", ctypes.c_uint8, 3),       # 3bit zero
        ("rcode", ctypes.c_uint8, 4),   # 4bit response code
        ("qdcount", ctypes.c_uint16),   # 16bit question count
        ("ancount", ctypes.c_uint16),   # 16bit answer count
        ("nscount", ctypes.c_uint16),   # 16bit count of the name server resource records
        ("arcount", ctypes.c_uint16)    # 16bit additional record count
    ]

    def __repr__(self):
        pp = PPrinter()
        pp.add(text=f"0x{self.id:04x}", bitlen=16)
        pp.add(text=f"{self.qr:01b}", bitlen=1)
        pp.add(text=f"0x{self.opcode:01x}", bitlen=4)
        pp.add(text=f"{self.aa:01b}", bitlen=1)
        pp.add(text=f"{self.tc:01b}", bitlen=1)
        pp.add(text=f"{self.rd:01b}", bitlen=1)
        pp.add(text=f"{self.ra:01b}", bitlen=1)
        pp.add(text=f"{self.z:01x}", bitlen=3)
        pp.add(text=f"0x{self.rcode:01x}", bitlen=4)
        pp.add(text=f"0x{self.qdcount:04x}", bitlen=16)
        pp.add(text=f"0x{self.ancount:04x}", bitlen=16)
        pp.add(text=f"0x{self.nscount:04x}", bitlen=16)
        pp.add(text=f"0x{self.arcount:04x}", bitlen=16)
        return str(pp)


class DNSQuestion:
    def __init__(self, domain, qtype=QTYPE.A, qclass=QCLASS.IN):
        self.qname = domain if isinstance(domain, DNSDomain) else DNSDomain(domain)
        self.qtype = qtype
        self.qclass = qclass

    def __bytes__(self):
        q = bytes(self.qname)
        q += self.qtype.to_bytes(2, byteorder="big")
        q += self.qclass.to_bytes(2, byteorder="big")
        return q

    def __len__(self):
        return len(self.qname) + 4

    def __repr__(self):
        pp = PPrinter(attach=True)
        pp.add(text=f"{self.qname}", bitlen=16, flex=True)
        pp.add(text=f"0x{self.qtype:04x}", bitlen=16)
        pp.add(text=f"0x{self.qclass:04x}", bitlen=16)
        return str(pp)

    @classmethod
    def from_rsp(cls, rsp, index=12) -> 'DNSQuestion':
        def to_int(data):
            return int.from_bytes(data, byteorder="big")

        domain = DNSDomain.to_domain(rsp, index)
        index += len(domain)
        return cls(domain=domain,
                   qtype=to_int(rsp[index:index + 2]),
                   qclass=to_int(rsp[index + 2:index + 4]))



import socket

from kydns.protocol import PPrinter
from kydns.kyd_models import DNSDomain, QTYPE, QCLASS


class DNSRecord:
    def __init__(self,
                 name: DNSDomain,
                 rtype: int = QTYPE.A,      # 16bit type code
                 rclass: int = QCLASS.IN,   # 16bit class code
                 ttl: int = 0,              # 32bit int, valid for, 0=don't cache
                 rdlength: int = 0,         # 32bit uint, rdata length in bytes
                 rdata: bytes = b"0000",    # answer, varies according to rtype and rclass
                 ans: str = "",             # rdata in a human readable form
                 ):
        self.name = name
        self.rtype = rtype
        self.rclass = rclass
        self.ttl = ttl
        self.rdlength = rdlength
        self.rdata = rdata
        self.ans = ans

    def __len__(self):
        return len(self.__bytes__())

    def __bytes__(self):
        ans = bytes(self.name)
        ans += self.rtype.to_bytes(2, byteorder="big")
        ans += self.rclass.to_bytes(2, byteorder="big")
        ans += self.ttl.to_bytes(4, byteorder="big")
        ans += self.rdlength.to_bytes(2, byteorder="big")
        ans += self.rdata
        return ans

    def __repr__(self):
        pp = PPrinter(attach=True)
        pp.add(text=f"{self.name}", bitlen=16, flex=True)
        pp.add(text=f"0x{self.rtype:04x}", bitlen=16)
        pp.add(text=f"0x{self.rclass:04x}", bitlen=16)
        pp.add(text=f"{self.ttl}", bitlen=32)
        pp.add(text=f"0x{self.rdlength:04x}", bitlen=16)
        pp.add(text=f"{self.ans}", bitlen=32, flex=True)
        return str(pp)

    @classmethod
    def from_rsp(cls, rtype: int, rsp: bytes, index: int):
        record_cls = RTYPE_MAPPER.get(rtype)
        if not record_cls:
            raise NotImplementedError(f"Unable to parse response for unsupported type '{rtype}'")

        domain = DNSDomain.to_domain(rsp, index)
        index += len(domain)

        rtype = to_int(rsp[index:index + 2])
        rclass = to_int(rsp[index + 2:index + 4])
        ttl = to_int(rsp[index + 4:index + 8])
        rdlength = to_int(rsp[index + 8:index + 10])
        rdata = rsp[index + 10:index + 10 + rdlength]
        ans = record_cls.to_ans(rsp, index + 10, rdlength)

        return RTYPE_MAPPER.get(rtype)(domain, rtype, rclass, ttl, rdlength, rdata, ans)


class ARecord(DNSRecord):
    @staticmethod
    def to_ans(rsp, rdata_index, rdlength):
        return socket.inet_ntop(socket.AF_INET, rsp[rdata_index:rdata_index + rdlength])


class AAAARecord(DNSRecord):
    @staticmethod
    def to_ans(rsp, rdata_index, rdlength):
        return socket.inet_ntop(socket.AF_INET6, rsp[rdata_index:rdata_index + rdlength])


class NSRecord(DNSRecord):
    @staticmethod
    def to_ans(rsp, rdata_index, rdlength):
        return DNSDomain.to_domain(rsp, rdata_index)


def to_int(data: bytes) -> int:
    return int.from_bytes(data, byteorder="big")


RTYPE_MAPPER = {
    QTYPE.A: ARecord,
    QTYPE.AAAA: AAAARecord,
    QTYPE.NS: NSRecord,
}



__version__ = "0.1.2"

import secrets
import socket

from kydns.kyd_exc import *
from kydns.kyd_models import DNSHeader, DNSQuestion
from kydns.kyd_records import DNSRecord

DNS_HEADER_LEN = 12


class Request:
    def __init__(self, domain, qtype=1, qclass=1, id=None,
                 qr=0, opcode=0, aa=0, tc=0, rd=1, ra=0, z=0, rcode=0,
                 qdcount=1, ancount=0, nscount=0, arcount=0):
        self.header = DNSHeader(id=id or int(secrets.token_hex(2), 16),
                                qr=qr, opcode=opcode, aa=aa, tc=tc, rd=rd, ra=ra, z=z, rcode=rcode,
                                qdcount=qdcount, ancount=ancount, nscount=nscount, arcount=arcount)
        self.question = DNSQuestion(domain, qtype, qclass)

    def __repr__(self):
        return str(self.header) + str(self.question)

    def __bytes__(self):
        return bytes(self.header) + bytes(self.question)

    def send(self, addr, timeout: int = 1) -> 'Response':
        sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        sock.settimeout(timeout)
        try:
            sock.sendto(bytes(self), addr)
            try:
                data, server = sock.recvfrom(4096)
            except socket.timeout:
                raise DNSResponseTimeout(f"Timed out after waiting for response from {addr} after {timeout}s")
        finally:
            sock.close()
        return self.to_rsp(rsp=data, rsp_server=server)

    def to_rsp(self, rsp: bytes, rsp_server: tuple = None) -> 'Response':
        obj = Response(rsp_server)
        obj.header = DNSHeader.from_buffer_copy(rsp[:DNS_HEADER_LEN])
        self.raise_on_err(obj.header)
        obj.question = DNSQuestion.from_rsp(rsp)
        index = DNS_HEADER_LEN + len(obj.question)
        for _ in range(obj.header.ancount):
            record = DNSRecord.from_rsp(obj.question.qtype, rsp, index)
            index += len(record)
            obj.add_record(record)
        return obj

    def raise_on_err(self, rsp_header) -> None:
        if self.header.id != rsp_header.id:
            raise DNSError(f"Response ID does not match: {self.header.id} != {rsp_header.id}")
        if rsp_header.rcode:
            if rsp_header.rcode == 1:
                raise DNSFormatError(f"Name server unable to interpret the query")
            if rsp_header.rcode == 2:
                raise DNSServerFailure()
            if rsp_header.rcode == 3:
                raise DNSNameError("Requested domain name does not exist")
            if rsp_header.rcode == 4:
                raise DNSNotImplemented("The name server does not support the requested query type")
            if rsp_header.rcode == 5:
                raise DNSRefused("The name server refuses to perform the specified operation")
        if rsp_header.qdcount > 1:
            raise DNSError(f"Responses with more than 1 question are unsupported")


class Response:
    def __init__(self, rsp_server: tuple = None):
        self.server = rsp_server or "NA"
        self.header = None
        self.question = None
        self.records = []

    def __repr__(self):
        return str(self.header) + str(self.question) + "".join(str(r) for r in self.records)

    def __bytes__(self):
        return bytes(self.header) + bytes(self.question) + b"".join(bytes(r) for r in self.records)

    def add_record(self, record):
        self.records.append(record)



class PPrinter:
    LINEWRAP = ["|", "+"]
    FLEXWRAP = ["/", "+"]
    FIELD_SEP = "|"

    def __init__(self, bits: int = 16, attach=False):
        self.bits = bits            # number of bits each line covers
        self.chars = bits * 2 - 1   # max number of characters per line
        self.curr_line = ""         # contains a line that wasn't filled by a single field
        self.free_bits = bits       # contains number of free bits remaining in such line
        self.fields = [] if attach else [self.get_top_numbers(), self.get_line_separator()]

    def get_top_numbers(self) -> str:
        """get heading numbers

        Returns:
            heading numbers lines
        """

        return "\n".join(["  " + "".join([" " * 19 + str(i + 1) for i in range(self.bits // 10)]),
                          " " + " ".join([str(i % 10) for i in range(self.bits)])])

    def get_line_separator(self) -> str:
        """get a separating line as a string

        Returns:
            separating line
        """

        return "+-" * self.bits + "+"

    def get_section_separator(self) -> str:
        """get a section separating linea as a string

         Returns:
             section separating line
         """

        return "##" * self.bits + "#"

    def is_line_full(self) -> bool:
        """check if current line is full

        Returns:
            True if current line is full
        """

        return self.free_bits == 0

    def is_new_line(self) -> bool:
        """check if current line is empty

        Returns:
            True if current line is empty
        """

        return self.free_bits == self.bits

    def can_fill_line(self, bitlen: int) -> bool:
        """check if field size may fill a single line fully

        Args:
            bitlen:         field length in bits

        Returns:
            True if field length is exactly total bits per line
        """

        return bitlen == self.bits

    def can_fill_lines(self, bitlen: int) -> bool:
        """check if field size may fill multiple lines fully

        Args:
            bitlen:         field length in bits

        Returns:
            True if field length is a multiple of total bits per line
        """

        return bitlen % self.bits == 0

    def add_singleline(self, text: str, flex: bool) -> None:
        """add a line that fits exactly on a single line

        Args:
            text:           field text
            flex:           flexible field len [default: False]
        """

        self.fields.append(self.wrap_line(self.center_text(text, self.bits), flex=flex))

    def add_multiline(self, text: str, bitlen: int, flex: bool) -> None:
        """add a line that fits exactly in multiple lines

        Args:
            text:           field text
            bitlen:         field length in bits
            flex:           flexible field len [default: False]
        """

        total_lines = (bitlen // self.bits) * 2 - 1
        mid_ind = total_lines // 2
        for i in range(total_lines):
            self.fields.append(self.wrap_line(self.center_text(text if i == mid_ind else "", self.bits),
                                              multiline_ind=i, flex=flex))

    def add_multifield_line(self) -> None:
        """add a line that's from multiple fields
        1. add current line
        2. reset free bits to max (self.bits)
        3. reset current line to empty
        """

        self.fields.append(self.wrap_line(self.curr_line))
        self.free_bits = self.bits
        self.curr_line = ""

    def add_field(self, text: str, bitlen: int) -> None:
        """adds a field that doesn't cover a full line, to the current line

        Args:
            text:           field text
            bitlen:         field length in bits
        """

        self.free_bits -= bitlen
        self.curr_line += self.center_text(text, bitlen)

    def append(self, text: str, bitlen: int) -> None:
        """append a field that doesn't cover a full line

        Args:
            text:           field text
            bitlen:         field length in bits

        Raises:
            NotImplementedError: if field is size is not supported.
                                 ie can't fit in one line
        """

        if bitlen <= self.free_bits:
            self.add_field(text, bitlen)
        else:
            raise NotImplementedError

        if self.is_line_full():
            self.add_multifield_line()
        else:
            self.curr_line = self.wrap_field(self.curr_line)

    def add(self, text: str, bitlen: int, flex: bool = False) -> None:
        """add a field to the protocol struct

        Args:
            text:           field text
            bitlen:         field length in bits
            flex:           flexible field len [default: False]

        Raises:
            NotImplementedError: if field is size is not supported.
                                 ie doesn't cover full line,
                                 nor adds up to full line
        """

        if self.is_new_line():
            if self.can_fill_line(bitlen):
                self.add_singleline(text, flex)
            elif self.can_fill_lines(bitlen):
                self.add_multiline(text, bitlen, flex)
            elif bitlen < self.bits:
                self.append(text, bitlen)
            else:
                raise NotImplementedError
        else:
            self.append(text, bitlen)

        if self.is_new_line():
            self.fields.append(self.get_line_separator())

    def __repr__(self):
        return "\n".join(self.fields) + "<--" + "\n"

    def center_text(self, text: str, bitlen: int) -> str:
        charlen = self.bitlen_to_charlen(bitlen)
        if len(text) > charlen:
            text = text[:charlen - 1] + "."
        return str.center(text, charlen)

    @staticmethod
    def wrap_field(field: str):
        """wrap a field by adding field separator

        Args:
            field:          str representing a field in line

        Returns:
            string of field appended with field separator
        """

        return field + PPrinter.FIELD_SEP

    @staticmethod
    def wrap_line(line: str, multiline_ind: int = 0, flex: bool = False) -> str:
        """once a line is ready to be printed, add starting and ending chars

        Args:
            line:           str representing a line to print
            multiline_ind:  line index in multiline fields [default: 0]
            flex:           flexible field len [default: False]

        Returns:
            str of ready for printing line
        """

        char = PPrinter.FLEXWRAP[multiline_ind % 2] if flex else PPrinter.LINEWRAP[multiline_ind % 2]
        return char + line + char

    @staticmethod
    def bitlen_to_charlen(bitlen: int) -> int:
        """get number of characters that covers the given number of bits

        Args:
            bitlen:         num bits

        Returns:
            num characters
        """

        return bitlen * 2 - 1



class DNSError(Exception):
    pass


# Request Errors

class DNSInvalidDomain(DNSError):
    pass


# Response Errors

class DNSFormatError(DNSError):
    pass


class DNSServerFailure(DNSError):
    pass


class DNSNameError(DNSError):
    pass


class DNSNotImplemented(DNSError):
    pass


class DNSRefused(DNSError):
    pass


class DNSResponseTimeout(DNSError):
    pass
